import io
import json
import pandas as pd
import pathlib
import tarfile
import sys
import time

from base64 import b64decode
from IPython.display import Image, HTML, display

from kubernetes import client, config
from kubernetes.stream import stream
from kubernetes.client.exceptions import ApiException


'''
POD_STATUS = {
    'Pending',
    'Running',
    'Succeeded',
    'Failed',
    'Unknown',
}
'''

def get_job_manifest(name, namespace, queue, image, cores, memory, num_gpu, node_selector):
    job_manifest = {
        'apiVersion': 'batch/v1',
        'kind': 'Job',
        'metadata': {
            'generateName': name,
            'labels': {
                'kueue.x-k8s.io/queue-name': queue,
            }
        },
        'spec': {
            'completions': 1,
            'template': {
                'metadata': {
                    'name': name,
                },
                'spec': {
                    'containers': [{
                        'name': 'notebook',
                        'image': image,
                        'resources': {
                            'requests': {
                               'cpu': cores,
                               'memory': memory,
                            },
                            'limits': {
                               'cpu': cores,
                               'memory': memory,
                               'nvidia.com/gpu': num_gpu
                            }
                        }
                    }],
                    'nodeSelector': {
                        'nvidia.com/gpu.product': node_selector,
                    },
                    'restartPolicy': 'Never',
                },
            }
        }
    }
    return job_manifest

def get_pod_manifest(name, image, cores, memory, num_gpu, node_selector):
    pod_manifest = {
        'apiVersion': 'v1',
        'kind': 'Pod',
        'metadata': {
            'name': name,
        },
        'spec': {
            'containers': [{
                'image': image,
                'name': 'notebook',
                'resources': {
                    'requests': {
                        'cpu': cores,
                        'memory': memory,
                    },
                    'limits': {
                        'cpu': cores,
                        'memory': memory,
                        'nvidia.com/gpu': num_gpu,
                    }
                }
            }],
            'nodeSelector': {
                'nvidia.com/gpu.product': node_selector,
            }
        }
    }
    return pod_manifest

def wait_for_pod(v1, name, ns, max_wait=60):
    resp = v1.read_namespaced_pod(name=name, namespace=ns)
    status = resp.status.phase
    i = 0
    while status == 'Pending' and i < max_wait:
        print('Waiting...')
        time.sleep(1)
        resp = v1.read_namespaced_pod(name=name, namespace=ns)
        status = resp.status.phase
        i += 1
        for cs in resp.status.container_statuses:
            try:
                if cs.state.waiting.message:
                    print(f'  message: {cs.state.waiting.message}')
                if cs.state.waiting.reason:
                    print(f'  reason: {cs.state.waiting.reason}')
            except:
                # no waiting message or reason
                pass
    if i >= max_wait:
        return 'Timeout'
    return status

def get_client_v1(args):
    kubeconfig = str(pathlib.Path(args.kubeconfig).expanduser().absolute())
    config.load_kube_config(config_file=kubeconfig)
    return client.CoreV1Api()

def get_batch_client_v1(args):
    kubeconfig = str(pathlib.Path(args.kubeconfig).expanduser().absolute())
    config.load_kube_config(config_file=kubeconfig)
    return client.BatchV1Api()

def connect_pod(name, args, launch=True):
    v1 = get_client_v1(args)
    batch_v1 = get_batch_client_v1(args)
    pod_name = name
    job_name = None
    if launch:
        job_manifest = get_job_manifest(
            name=name,
            namespace=args.namespace,
            queue=args.queue,
            image=args.image,
            cores=args.cores,
            memory=f'{args.memory}Gi',
            num_gpu=args.gpus,
            node_selector=args.gpu_product,
        )
        resp = batch_v1.create_namespaced_job(body=job_manifest, namespace=args.namespace)
        job_name = resp.metadata.name
        print(f'Waiting for job {job_name} to start ', end='')
        while resp.status.active is None or resp.status.active < 1:
            resp = batch_v1.read_namespaced_job(name=job_name, namespace=args.namespace)
            time.sleep(1)
            print('.', end='')
        try:
            pod_list = v1.list_namespaced_pod(args.namespace, label_selector=f'job-name={job_name}')
            pod_name = pod_list.items[0].metadata.name
            print(f'\nWaiting for pod: {pod_name}')
        except:
            # no pod although status is active
            print('Could not find pod', file=sys.stderr)
            return
    status = wait_for_pod(v1, pod_name, args.namespace)
    result =  {
        'success': False,
        'message': 'An unknown error occurred.',
        'pod_name': pod_name,
        'job_name': job_name,
        'client': v1,
        'batch_client': batch_v1,
    }
    if status == 'Running':
        # we're good
        result['success'] = True
        result['message'] = 'Container is running.'
    elif status == 'Timeout':
        result['message'] = 'Timed out while waiting for a running Pod.'
    elif status == 'Succeeded':
        result['message'] = 'All containers in the Pod have terminated in success, and will not be restarted.'
    elif status == 'Failed':
        result['message'] = 'All containers in the Pod have terminated, and at least one container has terminated in failure.'
    elif status == 'Unknown':
        result['message'] = 'For some reason the state of the Pod could not be obtained.'
    return result

def copy_file(kube_conn, namespace: str, pod_name: str, source_file: pathlib.Path, dest_path: str):
    buf = io.BytesIO()
    with tarfile.open(fileobj=buf, mode='w:tar') as tar:
        tar.add(source_file, arcname=pathlib.Path(dest_path).joinpath(source_file.name))
    commands = [buf.getvalue()]

    # Copying file
    exec_command = ['tar', 'xvf', '-', '-C', '/']
    resp = stream(kube_conn.connect_get_namespaced_pod_exec, pod_name, namespace,
                         command=exec_command,
                         stderr=True, stdin=True,
                         stdout=True, tty=False,
                         _preload_content=False)

    while resp.is_open():
        resp.update(timeout=1)
        if resp.peek_stdout():
            # print(f"STDOUT: {resp.read_stdout()}")
            # should be the file name
            pass
        if resp.peek_stderr():
            # do we need to close the stream?
            raise Exception(resp.read_stderr())
        if commands:
            c = commands.pop(0)
            resp.write_stdin(c)
        else:
            break
    resp.close()

def create_kernel(v1, pod_name, namespace, output_file='/tmp/kernel_id'):
    exec_command = [
        'python',
        '/tmp/create_kernel.py',
        output_file,
    ]
    kernel = stream(
        v1.connect_get_namespaced_pod_exec,
        pod_name,
        namespace,
        command=exec_command,
        stderr=True, stdin=False,
        stdout=True, tty=False
    )
    return kernel.split('\n')

def execute_python(v1, pod_name, namespace, kernel, session, code):
    exec_command = [
        'python', '/tmp/execute_in_kernel.py',
        kernel[0],
        kernel[1],
        kernel[2],
        session,
    ]
    resp = stream(v1.connect_get_namespaced_pod_exec,
                  pod_name,
                  namespace,
                  command=exec_command,
                  stderr=True, stdin=True,
                  stdout=True, tty=False,
                  _preload_content=False)

    codejson = json.dumps({'code': code})
    resp.write_stdin(f'{codejson}\n')
    result = None
    chunk = ''
    while resp.is_open():
        # print(resp)
        resp.update(timeout=1)
        if resp.peek_stdout():
            text = resp.read_stdout()
            for line in text.split('\n'):
                if line:
                    chunk += line
                    msg = None
                    try:
                        msg = json.loads(chunk)
                        # reset chunk to empty if previous messages combined to valid json
                        chunk = ''
                    except:
                        # might be a incomplete json so we continue to get more data
                        continue
                    try:
                        if msg['msg_type'] == 'execute_result' or msg['msg_type'] == 'display_data':
                            if 'data' in msg['content']:
                                data = msg['content']['data']
                                display(data, raw=True)
                        elif msg['msg_type'] == 'stream':
                            # text contains line breaks
                            if msg['content']['name'] == 'stderr':
                                print(msg['content']['text'], end='', file=sys.stderr)
                            else:
                                print(msg['content']['text'], end='')
                        elif msg['msg_type'] == 'error':
                            for tb in msg['content']['traceback']:
                                print(tb, file=sys.stderr)
                    except:
                        # import traceback
                        # traceback.print_exc()
                        # there's something else wrong so we discard previous data
                        chunk = ''
                        print(line)
        if resp.peek_stderr():
            print(resp.read_stderr(), file=sys.stderr)
    resp.close()
    return result

def delete_pod(v1, pod_name, namespace, max_wait=60):
    v1.delete_namespaced_pod(name=pod_name, namespace=namespace)
    for i in range(max_wait):
        time.sleep(1)
        try:
            resp = v1.read_namespaced_pod(name=pod_name, namespace=namespace)
            status = resp.status.phase
            print(f'Container status: {status}')
        except ApiException as exc:
            if exc.status == 404:
                # pod cannot be found
                print('Pod deleted.')
                break
            else:
                raise

def delete_job(batch_v1, job_name, namespace, max_wait=60):
    resp = batch_v1.delete_namespaced_job(
        name=job_name,
        namespace=namespace,
        body=client.V1DeleteOptions(
                propagation_policy='Foreground',
                grace_period_seconds=10),
    )
    for i in range(max_wait):
        time.sleep(1)
        try:
            resp = batch_v1.read_namespaced_job(name=job_name, namespace=namespace)
            print(f'Status active: {resp.status.active}')
        except ApiException as exc:
            if exc.status == 404:
                # pod cannot be found
                print('Job deleted.')
                break
            else:
                raise

def list_pods(v1, namespace):
    pod_list = v1.list_namespaced_pod(namespace)
    pods = []
    for pod in pod_list.items:
        pods.append([
            pod.metadata.name,
            pod.status.phase,
            pod.status.start_time,
        ])
    df = pd.DataFrame(
        pods,
        columns=['NAME', 'STATUS', 'START TIME'],
    )
    return df.set_index('NAME')

def list_jobs(batch_v1, namespace):
    job_list = batch_v1.list_namespaced_job(namespace)
    jobs = []
    for job in job_list.items:
        # print(job.status)
        completions = 0
        all = 0
        for c in job.status.conditions:
            all += 1
            if c.status:
                completions += 1
        jobs.append([
            job.metadata.name,
            f'{completions}/{all}',
            job.status.start_time,
        ])
    df = pd.DataFrame(
        jobs,
        columns=['NAME', 'COMPLETIONS', 'START TIME'],
    )
    return df.set_index('NAME')
